class Telemetry:
    def __init__(self):
        self.telemetry_revision = None
        self.time_on = None
        self.status_bits = None
        self.reserved5 = None
        self.software_revision = None
        self.reserved17 = None
        self.frame_counter = None
        self.frame_mean = None
        self.fpa_temp_counts = None
        self.fpa_temp = None
        self.reserved25 = None
        self.fpa_temp_last_ffc = None
        self.last_ffc_time = None
        self.ffc_imminent = False
        self.ffc_status = None
